from .output_handler import HandleOutput
import os
import json
import sys


class HandleConfigs:

    FALLBACK_CONFIGURATION = {
    "log_components": {
        "color": True,
        "timestamp": True,
        "timestamp_format": "%Y-%m-%d / %Hh-%Mm-%Ss",
        "message": True,
        "level": True
    },
    "error_map": {
        "INFO": {
            "color": "\\033[0;34m",
            "level": "INFO",
            "traceback": False
        },
        "DEBUG": {
            "color": "\\033[0;35m",
            "level": "DEBUG",
            "traceback": True
        },
        "ERROR": {
            "color": "\\033[0;31m",
            "level": "ERROR",
            "traceback": False
        },
        "WARNING": {
            "color": "\\033[1;33m",
            "level": "WARNING",
            "traceback": True
        },
        "CRITICAL": {
            "color": "\\033[1;35m",
            "level": "CRITICAL",
            "traceback": True
        },
        "DEFAULT" : {
            "color": "\033[1;33m",
            "level": "DEFAULT",
            "traceback": True
        }
    },
    "file_locations": {
        "log_file_path": "{CURRENT_PATH}"
    },
    "log_rotation_configs": {
        "log_rotation": False,
        "size_rotation": True,
        "max_file_size": 10,
        "time_rotation": False,
        "max_file_age": 1,
        "check_interval": 30
    },

    "batch_logging_configs" : {
        "batch_logging": True,
        "batch_size": 10,
        "flush_interval": 45,
        "check_interval": 10
    },

    "output_configs": {
        "terminal": True,
        "file": True
    }
}

    def __init__(self, config_path):

        if config_path is None:
            config_path = os.path.join(os.path.dirname(
                __file__), 'Logger_primary_config.JSON')

        HandleOutput.log_internal_message(
            f"Attempting to load configs from {config_path}")
        self.configs = self.get_config_map(config_path)
        HandleOutput.log_internal_message("Initialising configuration file")

    def get_valid_config(self):

        self.validate_config_types(self.configs)
        self.validate_log_path()
        return self.configs

    def get_config_map(self, config_path):

        try:

            with open(config_path, 'r') as config:

                HandleOutput.log_internal_message(
                    "Custom config file successfully loaded")
                return json.load(config)

        except (FileNotFoundError, PermissionError, json.JSONDecodeError):

            with open(config_path, 'w') as config_file:

                HandleOutput.log_internal_message(
                    f"Specified config file could not be found or accessed.\nAttempting to create a file with default configurations as 'Logger_primary_config.JSON' at {config_path}.")
                json.dump(HandleConfigs.FALLBACK_CONFIGURATION,
                          config_file, indent=4)

            return HandleConfigs.FALLBACK_CONFIGURATION

    def validate_config_types(self, configs=None):

        if configs is None:
            configs = self.configs

        FALLBACK = HandleConfigs.FALLBACK_CONFIGURATION

        for key, value in configs.items():

            if key not in FALLBACK:
                HandleOutput.log_internal_message(
                    "Specified config uses malformed or corrupted data. Using Fallback")
                self.configs = FALLBACK
                return

            fallback = FALLBACK[key]

            if isinstance(value, dict) and isinstance(fallback, dict):

                for subkey, subvalue in value.items():

                    if subkey not in fallback:
                        HandleOutput.log_internal_message(
                            "Specified config uses malformed or corrupted data. Using Fallback")
                        self.configs = FALLBACK
                        return

                    subfallback = FALLBACK[key][subkey]

                    if isinstance(fallback[subkey], dict):

                        if not isinstance(subvalue, type(subfallback)):
                            HandleOutput.log_internal_message(
                                f"Encountered TypeError. Expected {type(subfallback)} but found {type(subvalue)}")
                            configs[key][subkey] = subfallback
                            continue

                continue

            if not isinstance(value, type(fallback)):
                HandleOutput.log_internal_message(
                    f"Encountered TypeError. Expected {type(fallback)} but found {type(value)}")
                configs[key] = fallback
                continue

    def validate_log_path(self):

        path = self.configs["file_locations"]["log_file_path"]

        if path == "{CURRENT_PATH}":
            path = os.path.join(os.path.dirname(sys.argv[0]), 'log.txt')

        try:

            with open(path, 'a', encoding='utf-8'):

                self.configs["file_locations"]["log_file_path"] = path
                HandleOutput.log_internal_message(
                    f"Log file found or created. Using {path}.")

        except (FileNotFoundError, PermissionError) as error:

            self.configs["output_configs"]["file"] = False
            self.configs["log_rotation_configs"]["log_rotation"] = False
            HandleOutput.log_internal_message(
                f"Path [{path}] to the log file couldn't be found, or can't be accessed. Disabled file rotation and logging", exception_traceback=error)



import os
import shutil
import sys
from random import randint
from datetime import datetime, timedelta
from .output_handler import HandleOutput


class HandleRotation:

    def __init__(self, configs):

        self.file_path = configs["file_locations"]["log_file_path"]
        self.archive_dir = os.path.dirname(self.file_path)

        self.log_rotation = configs["log_rotation_configs"]

        self.max_size = self.log_rotation["max_file_size"] # bytes
        self.max_age = self.log_rotation["max_file_age"] # days

    def check_rotation_condition(self, full_path):

        file_size = os.path.getsize(full_path)
        file_age =  datetime.now() - datetime.fromtimestamp(os.path.getctime(full_path))

        return file_size >= self.max_size or file_age >= timedelta(days=self.max_age)

    def rotate_log_path(self):

        HandleOutput.log_internal_message("Running rotation check")

        if self.check_rotation_condition(self.file_path):

            try:

                self.full_rotation()

            except (FileNotFoundError, PermissionError) as error:

                HandleOutput.log_internal_message("An error occured whilst attempting log rotation.", exception_traceback=error)

    def rename_rotating_log(self):
        archive_name = f"log_archive_{datetime.now().strftime('%d_%m_%y %H_%M_%S')}_{randint(1000, 99999)}.txt"
        archive_path = os.path.join(self.archive_dir, archive_name)
        
        os.makedirs(self.archive_dir, exist_ok=True)

        shutil.copy2(self.file_path, archive_path)

        archive_size = (os.path.getsize(archive_path) / 1024) # size in Kilobytes
        archive_age = os.path.getmtime(archive_path) # date/time of last metadata change

        HandleOutput.log_internal_message(f"Log file Archived as {archive_name}.\nArchived file size: {archive_size:.4f} KB\nArchived file age: {archive_age}.")

    def create_fresh_log(self):

        with open(self.file_path, 'w'):
            HandleOutput.log_internal_message("Cleared log file")

    def full_rotation(self):

        self.rename_rotating_log()
        self.create_fresh_log()


import threading
import signal
import time
from rotation_handler import HandleRotation


class BgRotation:

    def __init__(self, configs):

        self.load_configs(configs)
        self._rotation_handler = HandleRotation(self.full_configs)

        self.signals = [signal.SIGINT, signal.SIGTERM]
        self.create_signal()

        self.loop_break_key = threading.Event()

        self.task = threading.Thread(target=self.bg_rotation_check, daemon=True)
        self.task.start()

    def load_configs(self, configs):

        try:

            self.full_configs = configs
            self.interval = configs["log_rotation_configs"]["check_interval"]

        except KeyError as key_error:
            raise KeyError("Could not find log rotation configs. Log rotation is disabled.") from key_error

    def bg_rotation_check(self):

        while not self.loop_break_key.is_set():
            self._rotation_handler.rotate_log_path()

            time.sleep(self.interval)

    def break_task(self, signum, frame):
        self.loop_break_key.set()

    def create_signal(self):

        for signal_type in self.signals:
            signal.signal(signal_type, self.break_task)



from .output_handler import HandleOutput
import time
import threading

class HandleBatching:

    def __init__(self, file, batch_config):

        self.file_path = file

        self.log_buffer_max_size = batch_config["batch_size"]
        self.flush_interval = batch_config["flush_interval"]

        self._lock = threading.Lock()

        self.last_flush_time = time.time()

    def flush(self, buffer, override= False):

        with self._lock:

            if buffer or override:

                with open(self.file_path, 'a', encoding='utf-8') as log_file:
                    log_file.write('\n'.join(buffer) + '\n')

                buffer.clear()
                self.last_flush_time = time.time()
                HandleOutput.log_internal_message(f"Flushed {self.log_buffer_max_size} logs")

    def check_batching_condition(self, buffer):

            if len(buffer) >= self.log_buffer_max_size \
                or time.time() - self.last_flush_time >= self.flush_interval:

                return True
            
            return False




from format_util_ import FormatComponents

class CompileLog:

    def __init__(self, error_map, log_components):

        self.error_map = error_map
        self.color = log_components["color"]
        self.timestamp = log_components["timestamp"]
        self.level = log_components["level"]
        self.message = log_components["message"]

    def build_log(self, message, error_level="DEFAULT", exception_traceback=None):
            
            error_level_info = self.error_map.get(error_level, self.error_map["DEFAULT"])

            final_log = [
                FormatComponents.get_color(
                    self.error_map, error_level_info) if self.color else "",
                FormatComponents.get_timestamp(
                    self.log_components) if self.timestamp else "",
                f"[{error_level}] " if self.level else "",
                f"{message}" if self.message else "",
                FormatComponents.get_traceback(
                    exception_traceback) if error_level_info["traceback"] and exception_traceback else ""
            ]

            return ''.join(final_log)


from datetime import datetime
from output_handler import HandleOutput
import traceback


class FormatComponents:

    @staticmethod
    def get_color(error_map, error_level):

        return error_map[error_level]["color"].encode().decode("unicode_escape") 
        #.encode(),decode("unicode_escape") is used here because of 
        # how JSON files store ANSI escape sequences

    @staticmethod
    def get_timestamp(log_components):

        return f"[{datetime.now().strftime(log_components.get('timestamp_format', r'%y_%m_%d %H_%M_%S'))}]\n"

    @staticmethod
    def get_traceback(error):

        if error is None:

            HandleOutput.log_internal_message("No error was provided")
            return ""

        if not isinstance(error, Exception):
            if callable(error):

                error = error("Error passed with no message")
            else:
                HandleOutput.log_internal_message(
                    "Invalid error type or message")
                return ""

        if not hasattr(error, '__traceback__'):

            return "\nNo traceback provided"

        return "\n\033[0m" + ''.join(traceback.format_exception(None, error, error.__traceback__))



import traceback
import queue
import time
import atexit
import threading
from .batching_handler import HandleBatching

class HandleOutput:

    def __init__(self, output_config, batch_config, file):
        self.terminal_output = output_config["terminal"]
        self.file_output = output_config["file"]
        self.batch_config_toggle = batch_config["batch_logging"]

        self.file_path = file


        self.batcher = HandleBatching(file, batch_config)

        self._exit_flag = False

        self.log_buffer = []

        self.log_queue = queue.Queue()

        self.bg_task = threading.Thread(target=self._process_logs, daemon=True)
        self.bg_task.start()

        

        atexit.register(self._flush_on_exit)

    def _flush_on_exit(self):
        self.batcher.flush(self.log_buffer, override=True)
        self._exit_flag = True
        self.log_queue.put(None)

        self.bg_task.join()

    def _process_logs(self):
        
        while not self._exit_flag:
            message = self.log_queue.get(timeout=15)
            if message is None:
                break
            self.handle_log_output(message)

    def handle_log_output(self, message):

        if self.batch_config_toggle:

            if self.batcher.check_batching_condition(self.log_buffer):

                self.batcher.flush(self.log_buffer)

            else:

                    self.log_buffer.append(message)
        else:

            if self.terminal_output:

                print(message, flush=True)

            if self.file_output:

                    with open(self.file_path, 'a', encoding='utf-8') as log_file:
                        log_file.write(f"{message}\n")




    @staticmethod
    def log_internal_message(message, exception_traceback=None):

        print(f"\033[1;37m[INTERNAL] {message}")

        if exception_traceback:

            print(
                f"{traceback.format_exception(None, exception_traceback, exception_traceback.__traceback__)}")



from setuptools import setup, find_packages

setup(
    name='pylgx',  # Name of your package
    version='1.2.0',  # Starting version
    author='Kyroneff',  # Your name or organization
    author_email='Kyronfroom57@gmail.com',  # Your email
    description='A lightweight logging library for Python with features like log rotation and configurable error levels.',
    long_description=open('README.md').read(),  # Ensure you have a README.md
    long_description_content_type='text/markdown',  # Type of the README file
    url='https://github.com/KyronEff/Logging',  # Your project's URL
    packages=find_packages(),  # Automatically find and include all submodules
    classifiers=[  # Classifiers help others find your package
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',  # Minimum Python version
    install_requires=[  # List your dependencies here, if any
        # 'some_package>=1.0.0',
    ],
    include_package_data=True,  # Include any files listed in MANIFEST.in
    entry_points={  # If your package includes command-line tools
        'console_scripts': [
            'logger-cli=logger.cli:main',  # Example CLI command
        ],
    },
)



from .config_handler import HandleConfigs
from .log_compiler import CompileLog
from .output_handler import HandleOutput
from .bg_rotation import BgRotation

class Logger:

 # initialization function
    _compiler = None
    _output = None

    def __init__(self, config_path=None):

        self.configs = HandleConfigs(config_path).get_valid_config()

        if not Logger._compiler:
            Logger._compiler = CompileLog(
                self.configs["error_map"], self.configs["log_components"])
        if not Logger._output:
            Logger._output = HandleOutput(self.configs["output_configs"], self.configs["batch_logging_configs"], self.configs["file_locations"]["log_file_path"])

        self._compiler = Logger._compiler
        self._output = Logger._output

        if self.configs["log_rotation_configs"]["log_rotation"]:
            rotation_task = BgRotation(self.configs)


    def log(self, message, level, exception_traceback=None):

        message = self._compiler.build_log(message, level, exception_traceback)
        self._output.log_queue.put(message)

logx = Logger()

